# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import json

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak
from awkward._nplikes.shape import unknown_length

form_dict = {
    "class": "RecordArray",
    "fields": ["x", "y"],
    "contents": [
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "parameters": {},
                "form_key": "x.list.content",
            },
            "parameters": {},
            "form_key": "x.list.offsets",
        },
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "parameters": {},
                "form_key": "y.list.content",
            },
            "parameters": {},
            "form_key": "y.list.offsets",
        },
    ],
    "parameters": {},
}


def test_layout():
    form = ak.forms.from_dict(form_dict)
    layout, report = ak.typetracer.typetracer_with_report(form)
    assert isinstance(layout, ak.contents.Content)
    array = ak.Array(layout)

    y = array.y
    assert len(report.data_touched) == 0
    assert len(report.shape_touched) == 0

    ak.sum(y)
    assert set(report.data_touched) == {"y.list.offsets", "y.list.content"}
    assert set(report.shape_touched) == {"y.list.offsets", "y.list.content"}


def test_layout_highlevel_false():
    form = ak.forms.from_dict(form_dict)
    layout, report = ak.typetracer.typetracer_with_report(form, highlevel=False)
    assert isinstance(layout, ak.contents.Content)
    array = ak.Array(layout)

    y = array.y
    assert len(report.data_touched) == 0
    assert len(report.shape_touched) == 0

    ak.sum(y)
    assert set(report.data_touched) == {"y.list.offsets", "y.list.content"}
    assert set(report.shape_touched) == {"y.list.offsets", "y.list.content"}


def test_array_highlevel_true():
    form = ak.forms.from_dict(form_dict)
    array, report = ak.typetracer.typetracer_with_report(form, highlevel=True)
    assert isinstance(array, ak.Array)

    y = array.y
    assert len(report.data_touched) == 0
    assert len(report.shape_touched) == 0

    ak.sum(y)
    assert set(report.data_touched) == {"y.list.offsets", "y.list.content"}
    assert set(report.shape_touched) == {"y.list.offsets", "y.list.content"}


def test_form_dict():
    layout, report = ak.typetracer.typetracer_with_report(form_dict)
    assert isinstance(layout, ak.contents.Content)
    array = ak.Array(layout)

    y = array.y
    assert len(report.data_touched) == 0
    assert len(report.shape_touched) == 0

    ak.sum(y)
    assert set(report.data_touched) == {"y.list.offsets", "y.list.content"}
    assert set(report.shape_touched) == {"y.list.offsets", "y.list.content"}


def test_form_str():
    layout, report = ak.typetracer.typetracer_with_report(json.dumps(form_dict))
    assert isinstance(layout, ak.contents.Content)
    array = ak.Array(layout)

    y = array.y
    assert len(report.data_touched) == 0
    assert len(report.shape_touched) == 0

    ak.sum(y)
    assert set(report.data_touched) == {"y.list.offsets", "y.list.content"}
    assert set(report.shape_touched) == {"y.list.offsets", "y.list.content"}


def test_typetracer_from_form_highlevel_false():
    layout = ak.typetracer.typetracer_from_form(form_dict, highlevel=False)
    assert isinstance(layout, ak.contents.Content)
    assert layout.backend.name == "typetracer"
    assert layout.length is unknown_length


def test_typetracer_from_form_highlevel_true():
    array = ak.typetracer.typetracer_from_form(form_dict, highlevel=True)
    assert isinstance(array, ak.Array)
    assert ak.backend(array) == "typetracer"
    assert array.layout.length is unknown_length



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from datetime import datetime

import numpy as np
import pytest

import awkward as ak


def test_strings():
    assert ak.to_layout("hello").is_equal_to(
        ak.contents.NumpyArray(
            np.array([104, 101, 108, 108, 111], dtype=np.uint8),
            parameters={"__array__": "char"},
        )
    )
    assert ak.to_layout("hello", string_policy="promote").is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index64([0, 5]),
            ak.contents.NumpyArray(
                np.array([104, 101, 108, 108, 111], dtype=np.uint8),
                parameters={"__array__": "char"},
            ),
            parameters={"__array__": "string"},
        )
    )
    assert ak.to_layout("hello", string_policy="pass-through") == "hello"
    with pytest.raises(TypeError, match=r"string conversion/promotion is disabled"):
        ak.to_layout("hello", string_policy="error")


def test_primitives():
    with pytest.raises(TypeError, match=r"scalar conversion/promotion is disabled"):
        ak.to_layout(1j, primitive_policy="error")
    with pytest.raises(TypeError, match=r"scalar conversion/promotion is disabled"):
        ak.to_layout(False, primitive_policy="error")
    with pytest.raises(TypeError, match=r"scalar conversion/promotion is disabled"):
        ak.to_layout(datetime.now(), primitive_policy="error")

    with pytest.raises(TypeError, match=r"scalar conversion/promotion is disabled"):
        ak.to_layout(1, primitive_policy="error")
    assert ak.to_layout(1).is_equal_to(
        ak.contents.NumpyArray(np.array([1], dtype=np.int64))
    )
    out = ak.to_layout(1, primitive_policy="pass-through")
    assert isinstance(out, int) and out == 1

    with pytest.raises(TypeError, match=r"scalar conversion/promotion is disabled"):
        ak.to_layout(4.0, primitive_policy="error")
    assert ak.to_layout(4.0).is_equal_to(
        ak.contents.NumpyArray(np.array([4.0], dtype=np.float64))
    )
    out = ak.to_layout(4.0, primitive_policy="pass-through")
    assert isinstance(out, float) and out == 4.0


def test_with_field():
    result = ak.with_field([{"x": 1}], "hello", "x", highlevel=False)
    assert result.is_equal_to(
        ak.contents.RecordArray(
            [
                ak.contents.ListOffsetArray(
                    ak.index.Index64([0, 5]),
                    ak.contents.NumpyArray(
                        np.array([104, 101, 108, 108, 111], dtype=np.uint8),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string"},
                )
            ],
            ["x"],
        )
    )


def test_fill_none():
    array = ak.Array([None, 1, 2, None])
    result = ak.fill_none(array, "hello world!")
    assert result.layout.is_equal_to(
        ak.contents.UnionArray(
            ak.index.Index8([1, 0, 0, 1]),
            ak.index.Index64([0, 0, 1, 0]),
            [
                ak.contents.NumpyArray(np.array([1, 2], dtype=np.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64([0, 12]),
                    ak.contents.NumpyArray(
                        np.array(
                            [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100, 33],
                            dtype=np.uint8,
                        ),
                        parameters={"__array__": "char"},
                    ),
                    parameters={"__array__": "string"},
                ),
            ],
        )
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._typing import Protocol, Self


class PrivateSingleton(Protocol):
    _instance: Self

    @classmethod
    def _new(cls) -> Self:
        if hasattr(cls, "_instance"):
            raise RuntimeError(
                f"internal_error: singleton {cls.__name__} was already instantiated"
            )

        self = super().__new__(cls)
        self.__init__()
        cls._instance = self

        return self

    @classmethod
    def _ensure_instance(cls):
        try:
            return cls._instance
        except AttributeError:
            return cls._new()

    def __new__(cls, *args, **kwargs):
        raise TypeError(
            f"internal_error: {cls.__name__} class should never be directly instantiated."
        )

    def __reduce__(self):
        return type(self)._ensure_instance, ()

    @classmethod
    def _reduce_constructor(cls) -> Self:
        return cls._instance


class PublicSingleton(PrivateSingleton, Protocol):
    @classmethod
    def instance(cls) -> Self:
        return cls._ensure_instance()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from enum import IntEnum

__all__ = ("DLPackDevice",)


class DLPackDevice(IntEnum):
    CPU = 1  # CPU
    CUDA = 2  # GPU
    CUDA_PINNED = 3  # GPU & CPU
    OPENCL = 4  # UNSUPPORTED
    VULKAN = 7  # UNSUPPORTED
    METAL = 8  # UNSUPPORTED
    VPI = 9  # UNSUPPORTED
    ROCM = 10  # GPU
    ROCM_PINNED = 11  # GPU & CPU
    CUDA_MANAGED = 13  # GPU & CPU



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test():
    array = ak.Array([{"x": [[1, 2, 3], [4, 5, 6]]}])
    assert array[:, []].fields == ["x"]
    assert ak.to_list(ak.Array([{"x": [[1, 2, 3], [4, 5, 6]]}])[:, []]) == [{"x": []}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from awkward._connect.dlpack import DLPackDevice
from awkward._dispatch import high_level_function
from awkward._layout import from_arraylib, wrap_layout
from awkward._nplikes.cupy import Cupy
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpy_like import NumpyLike

__all__ = ("from_dlpack",)


@high_level_function()
def from_dlpack(
    array,
    *,
    prefer_cpu=True,
    regulararray=False,
    highlevel=True,
    behavior=None,
    attrs=None,
):
    """
    Args:
        array (cp.ndarray): The DLPack-supporting array to convert into an
            Awkward Array.
        prefer_cpu (bool): If True, and the array device supports both CPU and
            GPU backends, prefer the CPU; otherwise, prefer the GPU.
        regulararray (bool): If True and the array is multidimensional,
            the dimensions are represented by nested #ak.contents.RegularArray
            nodes; if False and the array is multidimensional, the dimensions
            are represented by a multivalued #ak.contents.NumpyArray.shape.
            If the array is one-dimensional, this has no effect.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Converts a DLPack-aware array into an Awkward Array.

    The resulting layout may involve the following #ak.contents.Content types
    (only):

    * #ak.contents.NumpyArray
    * #ak.contents.RegularArray if `regulararray=True`.
    """
    try:
        dlpack_info_func = array.__dlpack_device__
    except AttributeError as err:
        raise TypeError(
            f"Expected an object that implements the DLPack protocol, received {type(array)}"
        ) from err
    device_type, device_id = dlpack_info_func()

    # Only a subset of known devices are supported.
    nplike: NumpyLike
    if device_type == DLPackDevice.CPU:
        nplike = Numpy.instance()
    elif device_type == DLPackDevice.CUDA:
        nplike = Cupy.instance()
    elif device_type == DLPackDevice.CUDA_PINNED:
        # TODO: this should support GPU
        # nplike = (Numpy if prefer_cpu else Cupy).instance()
        nplike = Numpy.instance()
    elif device_type == DLPackDevice.ROCM:
        nplike = Cupy.instance()
    elif device_type == DLPackDevice.ROCM_PINNED:
        nplike = (Numpy if prefer_cpu else Cupy).instance()
    elif device_type == DLPackDevice.CUDA_MANAGED:
        nplike = (Numpy if prefer_cpu else Cupy).instance()
    else:
        raise AssertionError

    array = nplike.from_dlpack(array)
    return wrap_layout(
        from_arraylib(array, regulararray, False),
        highlevel=highlevel,
        behavior=behavior,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.Array([[1, 2, 3, 5, 7, 8, 10], [100000, 20000, 301]])
    quotient, remainder = np.divmod(array, 2)
    assert ak.almost_equal(quotient, [[0, 1, 1, 2, 3, 4, 5], [50000, 10000, 150]])
    assert ak.almost_equal(remainder, [[1, 0, 1, 1, 1, 0, 0], [0, 0, 1]])



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest
from packaging.version import parse as parse_version

import awkward as ak

if parse_version(np.__version__) < parse_version("1.23.0"):
    pytest.skip(
        "NumPy 1.23 or greater is required for DLPack testing", allow_module_level=True
    )


def test_from_dlpack_numpy():
    np_array = np.arange(2 * 3 * 4 * 5).reshape(2, 3, 4, 5)
    array = ak.from_dlpack(np_array)
    np_from_ak = ak.to_numpy(array)
    assert np.shares_memory(np_array, np_from_ak)


def test_from_dlpack_cupy():
    # This test only checks cupy usage, it doesn't explicitly test GPU & CPU
    cp = pytest.importorskip("cupy")
    cp_array = cp.arange(2 * 3 * 4 * 5).reshape(2, 3, 4, 5)
    array = ak.from_dlpack(cp_array)
    cp_from_ak = ak.to_cupy(array)
    assert cp.shares_memory(cp_array, cp_from_ak)


class DLPackOf:
    def __init__(self, array):
        self._array = array

    def __dlpack__(self, stream=None):
        if stream is None:
            return self._array.__dlpack__()
        else:
            return self._array.__dlpack__(stream)

    def __dlpack_device__(self):
        return self._array.__dlpack_device__()


def test_to_layout():
    np_array = np.arange(2 * 3 * 4 * 5).reshape(2, 3, 4, 5)
    dlpack_array = DLPackOf(np_array)
    layout = ak.to_layout(dlpack_array)
    assert layout.is_numpy

    np_from_ak = ak.to_numpy(layout)
    assert np.shares_memory(np_array, np_from_ak)


def test_invalid_argument():
    with pytest.raises(
        TypeError, match=r"Expected an object that implements the DLPack protocol"
    ):
        ak.from_dlpack([1, 2, 3])



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    layout = ak.to_layout(["hi", "bye"])

    def as_32(index):
        return ak.index.Index(index.nplike.astype(index.data, np.int32))

    layout = ak.contents.ListArray(
        as_32(layout.starts),
        as_32(layout.stops),
        layout.content,
        parameters=layout._parameters,
    )
    assert ak.pad_none(layout, 3, axis=0).tolist() == ["hi", "bye", None]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

from collections.abc import Mapping

from awkward._typing import Any, JSONMapping


def attrs_of_obj(obj, attrs: Mapping | None = None) -> Mapping | None:
    from awkward.highlevel import Array, ArrayBuilder, Record

    if attrs is not None:
        return attrs
    elif isinstance(obj, (Array, Record, ArrayBuilder)):
        return obj._attrs
    else:
        return None


def attrs_of(*arrays, attrs: Mapping | None = None) -> Mapping:
    # An explicit 'attrs' always wins.
    if attrs is not None:
        return attrs

    copied = False
    for x in reversed(arrays):
        x_attrs = attrs_of_obj(x)
        if x_attrs is None:
            continue
        if attrs is None:
            attrs = x_attrs
        elif attrs is x_attrs:
            pass
        elif not copied:
            attrs = dict(attrs)
            attrs.update(x_attrs)
            copied = True
        else:
            attrs.update(x_attrs)
    return attrs


def without_transient_attrs(attrs: dict[str, Any]) -> JSONMapping:
    return {k: v for k, v in attrs.items() if not k.startswith("@")}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest
from packaging.version import parse as parse_version

import awkward as ak

pyarrow = pytest.importorskip("pyarrow")
pyarrow_types = pytest.importorskip("pyarrow.types")


def test_simple_conversion():
    array = ak.Array(
        ak.contents.RecordArray(
            [
                ak.to_layout(np.arange(6, dtype=np.uint8)),
                ak.to_layout(np.arange(6, dtype=np.float64)),
            ],
            ["x", "y"],
        )
    )
    arrow_array = pyarrow.array(array)

    assert arrow_array.tolist() == array.to_list()


@pytest.mark.skipif(
    parse_version(pyarrow.__version__) < parse_version("12.0.0"),
    reason="pyarrow >= 12.0.0 required for casting test",
)
def test_type_cast():
    array = ak.mask(np.array([1, 2, 3], dtype=np.uint8), [True, False, False])
    arrow_array = pyarrow.array(array, type=pyarrow.float64())
    assert pyarrow_types.is_float64(arrow_array.type)

    arrow_array = pyarrow.array(array)
    assert pyarrow_types.is_uint8(arrow_array.type)


def test_zero_copy():
    np_array = np.array([1, 2, 3], dtype=np.uint32)
    ak_array = ak.from_numpy(np_array)
    arrow_array = pyarrow.array(ak_array)
    assert pyarrow_types.is_uint32(arrow_array.type)
    assert np.shares_memory(np_array, arrow_array.to_numpy())



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_ByteMaskedArray_getitem_nextcarry_64_1():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    mask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_getitem_nextcarry', cupy.int64, cupy.int8]
    funcC(tocarry, mask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 1, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_ByteMaskedArray_getitem_nextcarry_64_2():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    mask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = False
    funcC = cupy_backend['awkward_ByteMaskedArray_getitem_nextcarry', cupy.int64, cupy.int8]
    funcC(tocarry, mask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 1, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_ByteMaskedArray_getitem_nextcarry_64_3():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    mask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_getitem_nextcarry', cupy.int64, cupy.int8]
    funcC(tocarry, mask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 1, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))

def test_cuda_awkward_ByteMaskedArray_getitem_nextcarry_64_4():
    tocarry = cupy.array([123, 123, 123], dtype=cupy.int64)
    mask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_getitem_nextcarry', cupy.int64, cupy.int8]
    funcC(tocarry, mask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tocarry = [0, 1, 2]
    assert cupy.array_equal(tocarry[:len(pytest_tocarry)], cupy.array(pytest_tocarry))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

pytest.importorskip("pyarrow")


def test_to_categorical_nested():
    array = ak.Array([["one", "two", "three"], [], ["one", "two"], ["three"]])
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.str.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert ak.to_list(array) == categorical.to_list()
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical():
    array = ak.Array(
        ["one", "two", "three", "one", "two", "three", "one", "two", "three"]
    )
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.str.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert ak.to_list(array) == categorical.to_list()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical_none():
    array = ak.Array(
        [
            "one",
            "two",
            "three",
            None,
            "one",
            "two",
            "three",
            None,
            "one",
            "two",
            "three",
            None,
        ]
    )
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.str.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert ak.to_list(array) == categorical.to_list()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical_masked():
    content = ak.Array(
        [
            "one",
            "two",
            "three",
            "one",
            "one",
            "two",
            "three",
            "two",
            "one",
            "two",
            "three",
            "three",
        ]
    ).layout
    mask = ak.index.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
            ]
        )
    )
    array = ak.Array(ak.contents.ByteMaskedArray(mask, content, valid_when=False))
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.str.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert ak.to_list(array) == categorical.to_list()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical_masked_again():
    content = ak.Array(
        ["one", "two", "three", "one", "one", "two", "three", "two"]
    ).layout
    index = ak.index.Index64(
        np.array([0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3], dtype=np.int64)
    )
    indexedarray = ak.contents.IndexedArray(index, content)
    mask = ak.index.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
            ]
        )
    )
    array = ak.Array(
        ak.contents.ByteMaskedArray.simplified(mask, indexedarray, valid_when=False)
    )
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.str.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert ak.to_list(array) == categorical.to_list()
    assert ak.to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

@pytest.mark.skip(reason='Unable to generate any tests for kernel')
def test_cpuawkward_reduce_min_float64_float64_64_1():
    raise NotImplementedError('Unable to generate any tests for kernel')



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Literal

from awkward._backends.backend import Backend
from awkward._backends.dispatch import (
    common_backend,
    regularize_backend,
)
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of
from awkward._nplikes.dispatch import nplike_of_obj
from awkward._nplikes.jax import Jax
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpy_like import NumpyMetadata
from awkward._typing import TYPE_CHECKING, Any, NamedTuple, Self, TypeVar
from awkward.errors import AxisError

if TYPE_CHECKING:
    from awkward.contents import Content

np = NumpyMetadata.instance()
numpy = Numpy.instance()
numpy_backend = NumpyBackend.instance()


T = TypeVar("T")

if TYPE_CHECKING:
    from awkward.highlevel import Array
    from awkward.highlevel import Record as HighLevelRecord


class HighLevelMetadata(NamedTuple):
    backend: Backend
    attrs: Mapping | None
    behavior: Mapping | None


K = TypeVar("K")
V = TypeVar("V")


def merge_mappings(
    mappings: Sequence[Mapping[K, V]], default: Mapping[K, V] | None = None
) -> Mapping[K, V]:
    # TODO: add zero-copy optimisation
    if len(mappings) == 0:
        return default
    elif len(mappings) == 1:
        return mappings[0]
    else:
        return {k: v for mapping in mappings for k, v in mapping.items()}


class HighLevelContext:
    def __init__(
        self, behavior: Mapping | None = None, attrs: Mapping[str, Any] | None = None
    ):
        self._behavior = behavior
        self._attrs = attrs
        self._is_finalized = False

        self._attrs_from_objects = []
        self._behavior_from_objects = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finalize()

    def _ensure_finalized(self):
        if not self._is_finalized:
            raise RuntimeError("HighLevelContext has not yet been finalized")

    def _ensure_not_finalized(self):
        if self._is_finalized:
            raise RuntimeError("HighLevelContext has already been finalized")

    @property
    def attrs(self) -> Mapping[str, Any] | None:
        self._ensure_finalized()
        return self._attrs

    @property
    def behavior(self) -> Mapping | None:
        self._ensure_finalized()
        return self._behavior

    def finalize(self) -> Self:
        self._ensure_not_finalized()

        if self._behavior is None:
            # TODO: cleaner reverse
            behavior = merge_mappings(self._behavior_from_objects[::-1], default=None)
        else:
            behavior = self._behavior

        if self._attrs is None:
            attrs = merge_mappings(self._attrs_from_objects[::-1], default=None)
        else:
            attrs = self._attrs

        self._attrs = attrs
        self._behavior = behavior
        self._is_finalized = True

        return self

    def update(self, obj: T) -> T:
        from awkward.highlevel import Array, ArrayBuilder, Record

        self._ensure_not_finalized()

        if isinstance(obj, (Array, Record, ArrayBuilder)):
            if obj._attrs is not None:
                self._attrs_from_objects.append(obj._attrs)

            if obj._behavior is not None:
                self._behavior_from_objects.append(obj._behavior)

        return obj

    def unwrap(
        self,
        obj: Any,
        *,
        allow_record: bool = True,
        allow_unknown: bool = False,
        none_policy: Literal["error", "promote", "pass-through"] = "error",
        primitive_policy: Literal["error", "promote", "pass-through"] = "promote",
        string_policy: Literal[
            "error", "promote", "pass-through", "as-characters"
        ] = "as-characters",
        use_from_iter: bool = True,
        regulararray: bool = True,
    ) -> Any:
        from awkward.operations.ak_to_layout import _impl as to_layout_impl

        self.update(obj)

        return to_layout_impl(
            obj,
            allow_record=allow_record,
            allow_unknown=allow_unknown,
            none_policy=none_policy,
            use_from_iter=use_from_iter,
            primitive_policy=primitive_policy,
            string_policy=string_policy,
            regulararray=regulararray,
        )

    def wrap(
        self, obj: Any, *, highlevel: bool = True, allow_other: bool = False
    ) -> Any:
        self._ensure_finalized()

        return wrap_layout(
            obj,
            highlevel=highlevel,
            attrs=self._attrs,
            behavior=self._behavior,
            allow_other=allow_other,
        )


def ensure_same_backend(
    *layouts: Any, default_backend: str | Backend = "cpu"
) -> list[Any]:
    """

    Returns:
        object:
    """
    backends: set[Backend] = {
        layout.backend for layout in layouts if hasattr(layout, "backend")
    }

    backend: Backend
    if len(backends) >= 1:
        backend = common_backend(backends)
    else:
        backend = regularize_backend(default_backend)

    return [
        layout.to_backend(backend) if hasattr(layout, "to_backend") else layout
        for layout in layouts
    ]


def wrap_layout(
    content: T,
    behavior: Mapping | None = None,
    highlevel: bool = True,
    like: Any = None,
    allow_other: bool = False,
    attrs: Mapping | None = None,
) -> T | Array | HighLevelRecord:
    import awkward.highlevel
    from awkward.contents import Content
    from awkward.record import Record

    assert isinstance(content, (Content, Record)) or allow_other
    assert behavior is None or isinstance(behavior, Mapping)
    assert isinstance(highlevel, bool)
    if highlevel:
        if like is not None and behavior is None:
            behavior = behavior_of(like)

        if isinstance(content, Content):
            return awkward.highlevel.Array(content, behavior=behavior, attrs=attrs)
        elif isinstance(content, Record):
            return awkward.highlevel.Record(content, behavior=behavior, attrs=attrs)
        elif allow_other:
            return content
        else:
            raise AssertionError

    elif isinstance(content, (Content, Record)) or allow_other:
        return content
    else:
        raise AssertionError


def maybe_highlevel_to_lowlevel(obj):
    """
    Args:
        obj: an object

    Calls #ak.to_layout and returns the result iff. the object is a high-level
    Awkward object, otherwise the object is returned as-is.

    This function should be removed once scalars are properly handled by `to_layout`.
    """
    import awkward.highlevel

    if isinstance(
        obj,
        (
            awkward.highlevel.Array,
            awkward.highlevel.Record,
            awkward.highlevel.ArrayBuilder,
        ),
    ):
        return awkward.to_layout(obj)
    else:
        return obj


def from_arraylib(array, regulararray, recordarray):
    from awkward.contents import (
        ByteMaskedArray,
        ListArray,
        NumpyArray,
        RecordArray,
        RegularArray,
        UnmaskedArray,
    )
    from awkward.index import Index8, Index64

    # overshadows global NumPy import for nplike-safety
    nplike = nplike_of_obj(array)

    def recurse(array, mask=None):
        cls = type(array)
        if Jax.is_tracer_type(cls):
            raise TypeError("Jax tracers cannot be used with `ak.from_arraylib`")

        if regulararray and len(array.shape) > 1:
            new_shape = (-1,) + array.shape[2:]
            return RegularArray(
                recurse(nplike.reshape(array, new_shape), mask),
                array.shape[1],
                array.shape[0],
            )

        if len(array.shape) == 0:
            array = nplike.reshape(array, (1,))

        if array.dtype.kind == "S":
            assert nplike is numpy
            asbytes = array.reshape(-1)
            itemsize = asbytes.dtype.itemsize
            starts = numpy.arange(0, len(asbytes) * itemsize, itemsize, dtype=np.int64)
            stops = numpy.add(starts, numpy.char.str_len(asbytes))
            data = ListArray(
                Index64(starts),
                Index64(stops),
                NumpyArray(
                    asbytes.view("u1"),
                    parameters={"__array__": "byte"},
                    backend=numpy_backend,
                ),
                parameters={"__array__": "bytestring"},
            )
            for i in range(len(array.shape) - 1, 0, -1):
                data = RegularArray(data, array.shape[i], array.shape[i - 1])

        elif array.dtype.kind == "U":
            assert nplike is numpy
            asbytes = numpy.char.encode(array.reshape(-1), "utf-8", "surrogateescape")
            itemsize = asbytes.dtype.itemsize
            starts = numpy.arange(0, len(asbytes) * itemsize, itemsize, dtype=np.int64)
            stops = numpy.add(starts, numpy.char.str_len(asbytes))
            data = ListArray(
                Index64(starts),
                Index64(stops),
                NumpyArray(
                    asbytes.view("u1"),
                    parameters={"__array__": "char"},
                    backend=numpy_backend,
                ),
                parameters={"__array__": "string"},
            )
            for i in range(len(array.shape) - 1, 0, -1):
                data = RegularArray(data, array.shape[i], array.shape[i - 1])

        else:
            data = NumpyArray(array)

        if mask is None:
            return data

        elif mask is False or (isinstance(mask, np.bool_) and not mask):
            # NumPy's MaskedArray with mask == False is an UnmaskedArray
            if len(array.shape) == 1:
                return UnmaskedArray(data)
            else:

                def attach(x):
                    if isinstance(x, NumpyArray):
                        return UnmaskedArray(x)
                    else:
                        return RegularArray(attach(x.content), x.size, len(x))

                return attach(data.to_RegularArray())

        else:
            # NumPy's MaskedArray is a ByteMaskedArray with valid_when=False
            return ByteMaskedArray(Index8(mask), data, valid_when=False)

    if array.dtype == np.dtype("O"):
        raise TypeError("Awkward Array does not support arrays with object dtypes.")

    if isinstance(array, numpy.ma.MaskedArray):
        mask = numpy.ma.getmask(array)
        array = numpy.ma.getdata(array)
        if isinstance(mask, np.ndarray) and len(mask.shape) > 1:
            regulararray = True
            mask = mask.reshape(-1)
    else:
        mask = None

    if not recordarray or array.dtype.names is None:
        layout = recurse(array, mask)

    else:
        contents = []
        for name in array.dtype.names:
            contents.append(recurse(array[name], mask))
        layout = RecordArray(contents, array.dtype.names)

    return layout


def maybe_posaxis(layout: Content, axis: int, depth: int) -> int | None:
    from awkward.record import Record

    if isinstance(layout, Record):
        if axis == 0:
            raise AxisError("Record type at axis=0 is a scalar, not an array")
        return maybe_posaxis(layout._array, axis, depth)

    if axis >= 0:
        return axis

    else:
        is_branching, additional_depth = layout.branch_depth
        if not is_branching:
            return axis + depth + additional_depth - 1
        else:
            return None



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test():
    nparray = ak.contents.NumpyArray(np.array([1, 2, 3], dtype=np.int64))
    listarray = ak.contents.ListArray(
        ak.index.Index64(np.array([0], dtype=np.int64)),
        ak.index.Index64(np.array([3], dtype=np.int64)),
        nparray,
    )
    indexedarray = ak.contents.IndexedArray(ak.index.Index64([]), listarray)

    cart = ak.operations.cartesian([indexedarray, indexedarray], nested=True)
    assert str(cart.type) == "0 * var * var * (int64, int64)"
    assert to_list(cart) == []

    cart = ak.operations.cartesian([indexedarray, indexedarray], nested=False)
    assert str(cart.type) == "0 * var * (int64, int64)"
    assert to_list(cart) == []



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import io
from datetime import datetime, timedelta

import numpy as np
import pytest

import awkward as ak


def test_precision():
    stream = io.StringIO()
    ak.Array([1.0, 2.3456789]).show(stream=stream, precision=1)
    assert stream.getvalue() == "[1,\n 2]\n"

    stream.seek(0)
    ak.Array([1.0, 2.3456789]).show(stream=stream, precision=7)
    assert stream.getvalue() == "[1,\n 2.345679]\n"


def test_float():
    stream = io.StringIO()
    ak.Array([1.0, 2.3456789]).show(
        stream=stream, formatter={"float": "<FLOAT {}>".format}
    )
    assert stream.getvalue() == "[<FLOAT 1.0>,\n <FLOAT 2.3456789>]\n"

    stream.seek(0)
    ak.Array([1.0, 2.3456789]).show(
        stream=stream, formatter={"float_kind": "<FLOAT {}>".format}
    )
    assert stream.getvalue() == "[<FLOAT 1.0>,\n <FLOAT 2.3456789>]\n"


@pytest.mark.skipif(
    not hasattr(np, "float128"), reason="Only for 128-float supporting platforms"
)
def test_longfloat():
    stream = io.StringIO()
    ak.values_astype([1.0, 2.3456789], "float128").show(
        stream=stream, formatter={"longfloat": "<FLOAT {}>".format}
    )
    assert stream.getvalue() == "[<FLOAT 1.0>,\n <FLOAT 2.3456789>]\n"

    stream.seek(0)
    ak.values_astype([1.0, 2.3456789], "float128").show(
        stream=stream, formatter={"float_kind": "<FLOAT {}>".format}
    )
    assert stream.getvalue() == "[<FLOAT 1.0>,\n <FLOAT 2.3456789>]\n"


def test_complex():
    stream = io.StringIO()
    ak.Array([1.0j, 2.3456789j + 2]).show(
        stream=stream, formatter={"complexfloat": "<COMPLEX {}>".format}
    )
    assert stream.getvalue() == "[<COMPLEX 1j>,\n <COMPLEX (2+2.3456789j)>]\n"

    stream.seek(0)
    ak.Array([1.0j, 2.3456789j + 2]).show(
        stream=stream, formatter={"complex_kind": "<COMPLEX {}>".format}
    )
    assert stream.getvalue() == "[<COMPLEX 1j>,\n <COMPLEX (2+2.3456789j)>]\n"


@pytest.mark.skipif(
    not hasattr(np, "complex256"), reason="Only for 256-complex supporting platforms"
)
def test_longcomplex():
    stream = io.StringIO()
    ak.values_astype([1.0j, 2.3456789j + 2], "complex256").show(
        stream=stream, formatter={"longcomplexfloat": "<COMPLEX {}>".format}
    )
    assert stream.getvalue() == "[<COMPLEX 1j>,\n <COMPLEX (2+2.3456789j)>]\n"

    stream.seek(0)
    ak.values_astype([1.0j, 2.3456789j + 2], "complex256").show(
        stream=stream, formatter={"complex_kind": "<COMPLEX {}>".format}
    )
    assert stream.getvalue() == "[<COMPLEX 1j>,\n <COMPLEX (2+2.3456789j)>]\n"


def test_int():
    stream = io.StringIO()
    ak.Array([1, 2]).show(stream=stream, formatter={"int": "<INT {}>".format})
    assert stream.getvalue() == "[<INT 1>,\n <INT 2>]\n"

    stream.seek(0)
    ak.Array([1, 2]).show(stream=stream, formatter={"int_kind": "<INT {}>".format})
    assert stream.getvalue() == "[<INT 1>,\n <INT 2>]\n"


def test_bool():
    stream = io.StringIO()
    ak.Array([True, False]).show(stream=stream, formatter={"bool": "<BOOL {}>".format})
    assert stream.getvalue() == "[<BOOL True>,\n <BOOL False>]\n"


def test_str():
    stream = io.StringIO()
    ak.Array(["hello", "world"]).show(
        stream=stream, formatter={"str": "<STRING {!r}>".format}
    )
    assert stream.getvalue() == "[<STRING 'hello'>,\n <STRING 'world'>]\n"

    stream.seek(0)
    ak.Array(["hello", "world"]).show(
        stream=stream, formatter={"str_kind": "<STRING {!r}>".format}
    )
    assert stream.getvalue() == "[<STRING 'hello'>,\n <STRING 'world'>]\n"


def test_bytes():
    stream = io.StringIO()
    ak.Array([b"hello", b"world"]).show(
        stream=stream, formatter={"bytes": "<BYTES {!r}>".format}
    )
    assert stream.getvalue() == "[<BYTES b'hello'>,\n <BYTES b'world'>]\n"

    stream.seek(0)
    ak.Array([b"hello", b"world"]).show(
        stream=stream, formatter={"str_kind": "<STRING {!r}>".format}
    )
    assert stream.getvalue() == "[<STRING b'hello'>,\n <STRING b'world'>]\n"


def test_datetime64():
    stream = io.StringIO()
    ak.Array([datetime(year=2023, month=1, day=1, hour=12, minute=1, second=30)]).show(
        stream=stream, formatter={"datetime": "<DT {}>".format}
    )
    assert stream.getvalue() == "[<DT 2023-01-01T12:01:30.000000>]\n"


def test_timedelta64():
    stream = io.StringIO()
    ak.Array([timedelta(days=1, hours=12, minutes=1, seconds=30)]).show(
        stream=stream, formatter={"datetime": "<TD {}>".format}
    )
    assert stream.getvalue() == "[129690000000 microseconds]\n"



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_ListArray32_validity_1():
    starts = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int32)
    stops = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int32)
    length = 3
    lencontent = 3
    funcC = cupy_backend['awkward_ListArray_validity', cupy.int32, cupy.int32]

def test_cuda_awkward_ListArray32_validity_2():
    starts = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int32)
    stops = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int32)
    length = 3
    lencontent = 3
    funcC = cupy_backend['awkward_ListArray_validity', cupy.int32, cupy.int32]

def test_cuda_awkward_ListArray32_validity_3():
    starts = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int32)
    stops = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int32)
    length = 3
    lencontent = 3
    funcC = cupy_backend['awkward_ListArray_validity', cupy.int32, cupy.int32]
    funcC(starts, stops, length, lencontent)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")

def test_cuda_awkward_ListArray32_validity_4():
    starts = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int32)
    stops = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int32)
    length = 3
    lencontent = 3
    funcC = cupy_backend['awkward_ListArray_validity', cupy.int32, cupy.int32]

def test_cuda_awkward_ListArray32_validity_5():
    starts = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    stops = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int32)
    length = 3
    lencontent = 3
    funcC = cupy_backend['awkward_ListArray_validity', cupy.int32, cupy.int32]
    funcC(starts, stops, length, lencontent)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")




# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_1():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_2():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_3():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_4():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_5():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_6():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = False
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_7():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = False
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [0, 0, 0]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_8():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = False
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_9():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = False
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_10():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = False
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [0, 0, 0]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_11():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_12():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_13():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_14():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_15():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_16():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_17():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_18():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_19():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_20():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_21():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [0, 0, 0]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_22():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_23():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [0, 0, 0]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_24():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [0, 0, 0]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))

def test_cuda_awkward_ByteMaskedArray_overlay_mask8_25():
    tomask = cupy.array([123, 123, 123], dtype=cupy.int8)
    theirmask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    mymask = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int8)
    length = 3
    validwhen = True
    funcC = cupy_backend['awkward_ByteMaskedArray_overlay_mask', cupy.int8, cupy.int8, cupy.int8]
    funcC(tomask, theirmask, mymask, length, validwhen)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_tomask = [1, 1, 1]
    assert cupy.array_equal(tomask[:len(pytest_tomask)], cupy.array(pytest_tomask))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak
from awkward.typetracer import typetracer_with_report

SOME_ATTRS = {"foo": "FOO"}


def test_typetracer_with_report():
    array = ak.zip(
        {
            "x": [[0.2, 0.3, 0.4], [1, 2, 3], [1, 1, 2]],
            "y": [[0.1, 0.1, 0.2], [3, 1, 2], [2, 1, 2]],
            "z": [[0.1, 0.1, 0.2], [3, 1, 2], [2, 1, 2]],
        }
    )
    layout = ak.to_layout(array)
    form = layout.form_with_key("node{id}")

    meta, report = typetracer_with_report(form, highlevel=True, attrs=SOME_ATTRS)
    assert meta.attrs is SOME_ATTRS

    meta, report = typetracer_with_report(form, highlevel=True, attrs=None)
    assert meta._attrs is None


@pytest.mark.parametrize(
    "function",
    [
        ak.typetracer.touch_data,
        ak.typetracer.length_zero_if_typetracer,
        ak.typetracer.length_one_if_typetracer,
    ],
)
def test_function(function):
    array = ak.zip(
        {
            "x": [[0.2, 0.3, 0.4], [1, 2, 3], [1, 1, 2]],
            "y": [[0.1, 0.1, 0.2], [3, 1, 2], [2, 1, 2]],
            "z": [[0.1, 0.1, 0.2], [3, 1, 2], [2, 1, 2]],
        }
    )
    assert function(array, attrs=SOME_ATTRS).attrs is SOME_ATTRS
    assert function(array)._attrs is None



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest  # noqa: F401

import awkward as ak
from awkward._nplikes.shape import unknown_length


def test():
    fromjson = {
        "class": "RecordArray",
        "fields": ["muon", "anindex"],
        "contents": [
            {
                "class": "ListOffsetArray",
                "offsets": "i64",
                "content": {
                    "class": "RecordArray",
                    "fields": ["pt", "eta", "phi", "crossref", "thing1"],
                    "contents": [
                        {
                            "class": "NumpyArray",
                            "primitive": "int64",
                            "inner_shape": [],
                            "parameters": {},
                            "form_key": "muon_pt!",
                        },
                        {
                            "class": "NumpyArray",
                            "primitive": "int64",
                            "inner_shape": [],
                            "parameters": {},
                            "form_key": "muon_eta!",
                        },
                        {
                            "class": "NumpyArray",
                            "primitive": "int64",
                            "inner_shape": [],
                            "parameters": {},
                            "form_key": "muon_phi!",
                        },
                        {
                            "class": "ListOffsetArray",
                            "offsets": "i64",
                            "content": {
                                "class": "NumpyArray",
                                "primitive": "int64",
                                "inner_shape": [],
                                "parameters": {},
                                "form_key": "muon_crossref_content!",
                            },
                            "parameters": {},
                            "form_key": "muon_crossref_index!",
                        },
                        {
                            "class": "NumpyArray",
                            "primitive": "int64",
                            "inner_shape": [],
                            "parameters": {},
                            "form_key": "muon_thing1!",
                        },
                    ],
                    "parameters": {},
                    "form_key": "muon_record!",
                },
                "parameters": {},
                "form_key": "muon_list_outer!",
            },
            {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "parameters": {},
                "form_key": "anindex!",
            },
        ],
        "parameters": {},
        "form_key": "outer!",
    }

    form = ak.forms.from_dict(fromjson)

    ttlayout = form.length_zero_array(highlevel=False).to_typetracer(forget_length=True)

    ttarray = ak.Array(ttlayout, behavior=ak.behavior)
    backend = ttlayout.backend

    result = ak.unflatten(ttarray.muon.pt, ttarray.anindex, axis=1)
    assert result.layout.is_equal_to(
        ak.contents.ListOffsetArray(
            ak.index.Index64(
                backend.index_nplike.empty(unknown_length, dtype=np.int64)
            ),
            ak.contents.ListOffsetArray(
                ak.index.Index64(
                    backend.index_nplike.empty(unknown_length, dtype=np.int64)
                ),
                ak.contents.NumpyArray(
                    backend.nplike.empty(unknown_length, dtype=np.int64)
                ),
            ),
        )
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pickle

import numpy as np

import awkward as ak
from awkward._nplikes.numpy import Numpy
from awkward.typetracer import PlaceholderArray

nplike = Numpy.instance()


def test_round_trip_placeholder():
    layout = ak.contents.RecordArray(
        [
            ak.contents.NumpyArray(
                PlaceholderArray(nplike, (4,), np.dtype(np.float64))
            ),
            ak.contents.NumpyArray([1, 2, 3, 4]),
        ],
        ["x", "y"],
    )

    form, length, container = ak.to_buffers(layout)
    result = ak.from_buffers(form, length, container, highlevel=False)

    assert isinstance(result.content("x").data, PlaceholderArray)

    assert result.content("x").data.shape == layout.content("x").data.shape
    assert result.content("x").data.dtype == layout.content("x").data.dtype


def test_pickle_nplike():
    assert pickle.loads(pickle.dumps(nplike)) is nplike



from __future__ import annotations

import awkward as ak


def test():
    array = ak.typetracer.typetracer_from_form(
        ak.forms.from_type(
            ak.types.from_datashape("{x: var * {y:int64}}", highlevel=False)
        )
    )
    sliced = array[0, "x", 0]
    assert sliced.type.is_equal_to(
        ak.types.ScalarType(
            ak.types.RecordType([ak.types.NumpyType("int64")], ["y"]), None
        )
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._backends.typetracer import TypeTracerBackend
from awkward._dispatch import high_level_function
from awkward._layout import HighLevelContext, ensure_same_backend

__all__ = ("join",)

typetracer = TypeTracerBackend.instance()


@high_level_function(module="ak.str")
def join(array, separator, *, highlevel=True, behavior=None, attrs=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        separator (str, bytes, or array of them to broadcast): separator to
            insert between strings. If array-like, `separator` is broadcast
            against `array`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.
        attrs (None or dict): Custom attributes for the output array, if
            high-level.

    Concatenate the strings in `array`. The `separator` is inserted between
    each string. If array-like, `separator` is broadcast against `array` which
    permits a unique separator for each list of strings in `array`.

    Note: this function does not raise an error if the `array` does not
    contain any string or bytestring data.

    Requires the pyarrow library and calls
    [pyarrow.compute.binary_join](https://arrow.apache.org/docs/python/generated/pyarrow.compute.binary_join.html).

    See also: #ak.str.join_element_wise.
    """
    # Dispatch
    yield array, separator

    # Implementation
    return _impl(array, separator, highlevel, behavior, attrs)


def _is_maybe_optional_list_of_string(layout):
    if layout.is_list and layout.parameter("__array__") in {"string", "bytestring"}:
        return True
    elif layout.is_option or layout.is_indexed:
        return _is_maybe_optional_list_of_string(layout.content)
    else:
        return False


def _impl(array, separator, highlevel, behavior, attrs):
    from awkward._connect.pyarrow import import_pyarrow_compute

    pc = import_pyarrow_compute("ak.str.join")

    from awkward.operations.str import _apply_through_arrow

    with HighLevelContext(behavior=behavior, attrs=attrs) as ctx:
        layout, maybe_separator_layout = ensure_same_backend(
            ctx.unwrap(array, allow_record=False, primitive_policy="error"),
            ctx.unwrap(
                separator,
                allow_record=False,
                allow_unknown=False,
                primitive_policy="error",
                string_policy="pass-through",
            ),
        )

    if isinstance(maybe_separator_layout, (bytes, str)):

        def apply_unary(layout, **kwargs):
            if not (
                layout.is_list and _is_maybe_optional_list_of_string(layout.content)
            ):
                return

            return _apply_through_arrow(
                pc.binary_join,
                # Arrow needs an option type here
                layout.copy(
                    content=ak.contents.UnmaskedArray.simplified(layout.content)
                ),
                maybe_separator_layout,
                # This kernel requires non-large string/bytestrings
                string_to32=True,
                bytestring_to32=True,
            )

        out = ak._do.recursively_apply(layout, apply_unary)
    else:

        def apply_binary(layouts, **kwargs):
            if not (
                layouts[0].is_list
                and _is_maybe_optional_list_of_string(layouts[0].content)
            ):
                return

            if not _is_maybe_optional_list_of_string(layouts[1]):
                raise TypeError(
                    f"`separator` must be a list of (possibly missing) strings, not {ak.type(layouts[1])}"
                )

            return (
                _apply_through_arrow(
                    pc.binary_join,
                    # Arrow needs an option type here
                    layouts[0].copy(
                        content=ak.contents.UnmaskedArray.simplified(layouts[0].content)
                    ),
                    layouts[1],
                    # This kernel requires non-large string/bytestrings
                    string_to32=True,
                    bytestring_to32=True,
                ),
            )

        (out,) = ak._broadcasting.broadcast_and_apply(
            (layout, maybe_separator_layout), apply_binary
        )

    return ctx.wrap(out, highlevel=highlevel)



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_1():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [0, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_2():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [2, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_3():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [3, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_4():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [4, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_5():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [0, 0, 0]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_6():
    toindex = cupy.array([123], dtype=cupy.int64)
    outerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_7():
    toindex = cupy.array([123], dtype=cupy.int64)
    outerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_8():
    toindex = cupy.array([123], dtype=cupy.int64)
    outerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_9():
    toindex = cupy.array([123], dtype=cupy.int64)
    outerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_10():
    toindex = cupy.array([123], dtype=cupy.int64)
    outerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    innerlength = 2
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_11():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    innerlength = 5
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [1, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_12():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    innerlength = 5
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [1, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_13():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    innerlength = 5
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [1, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_14():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    innerlength = 5
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [1, 1, 1]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArrayU32_simplify32_to64_15():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    outerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.uint32)
    outerlength = 3
    innerindex = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    innerlength = 5
    funcC = cupy_backend['awkward_IndexedArray_simplify', cupy.int64, cupy.uint32, cupy.int32]
    funcC(toindex, outerindex, outerlength, innerindex, innerlength)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [0, 0, 0]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest

import awkward as ak

numba = pytest.importorskip("numba")


@numba.njit
def func(array):
    return array


def test_ArrayBuilder_behavior():
    SOME_ATTRS = {"FOO": "BAR"}
    builder = ak.ArrayBuilder(behavior=SOME_ATTRS)

    assert builder.behavior is SOME_ATTRS
    assert func(builder).behavior == SOME_ATTRS


def test_ArrayBuilder_non_picklable_behavior():
    def make_add_xyr():
        def add_xyr(left, right):
            x = left.x + right.x
            y = left.y + right.y
            return ak.zip(
                {
                    "x": x,
                    "y": y,
                    "r": np.sqrt(x**2 + y**2),
                },
                with_name="xyr",
            )

        return add_xyr

    behavior = {(np.add, "xyr", "xyr"): make_add_xyr()}
    builder = ak.ArrayBuilder(behavior=behavior)
    result = func(builder)
    assert result.behavior.keys() == behavior.keys()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from itertools import product

import pytest

import awkward as ak

behavior_1 = {"foo": "bar"}
behavior_2 = {"baz": "bargh!"}
behavior = {**behavior_1, **behavior_2}


@pytest.mark.parametrize(
    ("func", "axis"),
    [
        pytest.param(ak.softmax, 0, marks=pytest.mark.xfail()),
        pytest.param(ak.std, 0, marks=pytest.mark.xfail()),
        pytest.param(ak.var, 0, marks=pytest.mark.xfail()),
        (ak.softmax, 1),
        (ak.std, 1),
        (ak.var, 1),
        *product(
            (
                ak.any,
                ak.min,
                ak.argmin,
                ak.sum,
                ak.ptp,
                ak.count_nonzero,
                lambda *args, **kwargs: ak.moment(*args, **kwargs, n=3),
                ak.argmax,
                ak.all,
                ak.mean,
                ak.max,
                ak.prod,
                ak.count,
            ),
            ([0, 1]),
        ),
    ],
)
def test_reducers(axis, func):
    assert isinstance(
        func([[1, 2, 3, 4], [5], [10]], axis=axis, highlevel=True), ak.Array
    )
    assert isinstance(
        func([[1, 2, 3, 4], [5], [10]], axis=axis, highlevel=False), ak.contents.Content
    )
    assert (
        func(
            ak.Array([[1, 2, 3, 4], [5], [10]], behavior=behavior_1),
            axis=axis,
            highlevel=True,
            behavior=behavior_2,
        ).behavior
        == behavior_2
    )
    assert (
        func(
            ak.Array([[1, 2, 3, 4], [5], [10]], behavior=behavior_1),
            axis=axis,
            highlevel=True,
        ).behavior
        == behavior_1
    )


@pytest.mark.parametrize("func", [ak.covar, ak.corr, ak.linear_fit])
def test_covar(func):
    assert isinstance(
        func(
            [[1, 2, 3, 4], [5], [10]],
            [[4, 4, 0, 2], [1], [10]],
            axis=-1,
            highlevel=True,
        ),
        ak.Array,
    )
    assert isinstance(
        func(
            [[1, 2, 3, 4], [5], [10]],
            [[4, 4, 0, 2], [1], [10]],
            axis=-1,
            highlevel=False,
        ),
        ak.contents.Content,
    )
    assert (
        func(
            ak.Array(
                [[1, 2, 3, 4], [5], [10]],
                behavior=behavior_1,
            ),
            [[4, 4, 0, 2], [1], [10]],
            axis=-1,
            highlevel=True,
            behavior=behavior_2,
        ).behavior
        == behavior_2
    )
    assert (
        func(
            [[1, 2, 3, 4], [5], [10]],
            ak.Array(
                [[4, 4, 0, 2], [1], [10]],
                behavior=behavior_1,
            ),
            axis=-1,
            highlevel=True,
            behavior=behavior_2,
        ).behavior
        == behavior_2
    )
    assert (
        func(
            ak.Array(
                [[1, 2, 3, 4], [5], [10]],
                behavior=behavior_1,
            ),
            [[4, 4, 0, 2], [1], [10]],
            axis=-1,
            highlevel=True,
        ).behavior
        == behavior_1
    )
    assert (
        func(
            [[1, 2, 3, 4], [5], [10]],
            ak.Array(
                [[4, 4, 0, 2], [1], [10]],
                behavior=behavior_1,
            ),
            axis=-1,
            highlevel=True,
        ).behavior
        == behavior_1
    )
    assert (
        func(
            [[1, 2, 3, 4], [5], [10]],
            [[4, 4, 0, 2], [1], [10]],
            weight=ak.Array(
                [[1, 2, 3, 2], [1], [1]],
                behavior=behavior_1,
            ),
            axis=-1,
            highlevel=True,
            behavior=behavior_2,
        ).behavior
        == behavior_2
    )
    assert (
        func(
            [[1, 2, 3, 4], [5], [10]],
            [[4, 4, 0, 2], [1], [10]],
            weight=ak.Array(
                [[1, 2, 3, 2], [1], [1]],
                behavior=behavior_1,
            ),
            axis=-1,
            highlevel=True,
        ).behavior
        == behavior_1
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._dispatch import high_level_function

__all__ = ("is_tuple",)


@high_level_function()
def is_tuple(array):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).

    If `array` is a record, this returns True if the record is a tuple.
    If `array` is an array, this returns True if the outermost record is a tuple.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array)


def _impl(array):
    layout = ak.operations.ak_to_layout._impl(
        array,
        allow_record=True,
        allow_unknown=False,
        none_policy="error",
        regulararray=True,
        use_from_iter=True,
        primitive_policy="error",
        string_policy="as-characters",
    )

    return layout.is_tuple



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_IndexedArray_fill_to64_count_1():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    toindexoffset = 0
    length = 3
    base = 3
    funcC = cupy_backend['awkward_IndexedArray_fill_count', cupy.int64]
    funcC(toindex, toindexoffset, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [3, 4, 5]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArray_fill_to64_count_2():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    toindexoffset = 0
    length = 3
    base = 3
    funcC = cupy_backend['awkward_IndexedArray_fill_count', cupy.int64]
    funcC(toindex, toindexoffset, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [3, 4, 5]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArray_fill_to64_count_3():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    toindexoffset = 0
    length = 3
    base = 3
    funcC = cupy_backend['awkward_IndexedArray_fill_count', cupy.int64]
    funcC(toindex, toindexoffset, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [3, 4, 5]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArray_fill_to64_count_4():
    toindex = cupy.array([123, 123, 123, 123], dtype=cupy.int64)
    toindexoffset = 1
    length = 3
    base = 3
    funcC = cupy_backend['awkward_IndexedArray_fill_count', cupy.int64]
    funcC(toindex, toindexoffset, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [123, 3, 4, 5]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))

def test_cuda_awkward_IndexedArray_fill_to64_count_5():
    toindex = cupy.array([123, 123, 123], dtype=cupy.int64)
    toindexoffset = 0
    length = 3
    base = 3
    funcC = cupy_backend['awkward_IndexedArray_fill_count', cupy.int64]
    funcC(toindex, toindexoffset, length, base)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_toindex = [3, 4, 5]
    assert cupy.array_equal(toindex[:len(pytest_toindex)], cupy.array(pytest_toindex))




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest  # noqa: F401

import awkward as ak

form_dict = {
    "class": "RecordArray",
    "fields": ["x"],
    "contents": [
        {
            "class": "ListArray",
            "starts": "i64",
            "stops": "i64",
            "content": {
                "class": "NumpyArray",
                "primitive": "int64",
                "inner_shape": [],
                "parameters": {},
                "form_key": "x.list.content",
            },
            "parameters": {},
            "form_key": "x.list",
        }
    ],
    "parameters": {},
}


def test_without_attribute():
    form = ak.forms.from_dict(form_dict)
    array, report = ak.typetracer.typetracer_with_report(
        form, buffer_key="{form_key}", highlevel=True
    )
    ak.typetracer.touch_data(array)
    assert set(report.data_touched) == {
        "x.list",
        "x.list.content",
    }


def test_with_attribute():
    form = ak.forms.from_dict(form_dict)
    array, report = ak.typetracer.typetracer_with_report(
        form, buffer_key="{form_key}-{attribute}", highlevel=True
    )
    ak.typetracer.touch_data(array)
    assert set(report.data_touched) == {
        "x.list-starts",
        "x.list-stops",
        "x.list.content-data",
    }



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak


def test_ByteMaskedArray():
    record = ak.zip(
        {
            "x": ak.mask(["foo", "bar", "world"], [True, True, False]),
            "y": ak.mask(["do", "re", "mi"], [False, True, True]),
        },
        optiontype_outside_record=True,
    )
    assert record.to_list() == [None, {"x": "bar", "y": "re"}, None]
    assert record.type == ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.RecordType(
                [
                    ak.types.ListType(
                        ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
                        parameters={"__array__": "string"},
                    ),
                    ak.types.ListType(
                        ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
                        parameters={"__array__": "string"},
                    ),
                ],
                ["x", "y"],
            )
        ),
        3,
        None,
    )


def test_IndexedOptionArray():
    record = ak.zip(
        {"x": ["foo", "bar", None], "y": [None, "re", "mi"]},
        optiontype_outside_record=True,
    )
    assert record.to_list() == [None, {"x": "bar", "y": "re"}, None]
    assert record.type == ak.types.ArrayType(
        ak.types.OptionType(
            ak.types.RecordType(
                [
                    ak.types.ListType(
                        ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
                        parameters={"__array__": "string"},
                    ),
                    ak.types.ListType(
                        ak.types.NumpyType("uint8", parameters={"__array__": "char"}),
                        parameters={"__array__": "string"},
                    ),
                ],
                ["x", "y"],
            )
        ),
        3,
        None,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import awkward as ak
from awkward._nplikes.array_like import ArrayLike
from awkward._nplikes.array_module import ArrayModuleNumpyLike
from awkward._nplikes.dispatch import register_nplike
from awkward._nplikes.numpy_like import UfuncLike
from awkward._typing import Final, cast


@register_nplike
class Jax(ArrayModuleNumpyLike):
    is_eager: Final = True
    supports_structured_dtypes: Final = False

    def __init__(self):
        jax = ak.jax.import_jax()
        self._module = jax.numpy

    def prepare_ufunc(self, ufunc: UfuncLike) -> UfuncLike:
        from awkward._connect.jax import get_jax_ufunc

        return get_jax_ufunc(ufunc)

    @property
    def ma(self):
        raise ValueError(
            "JAX arrays cannot have missing values until JAX implements "
            "numpy.ma.MaskedArray"
        )

    @property
    def char(self):
        raise ValueError(
            "JAX arrays cannot do string manipulations until JAX implements "
            "numpy.char"
        )

    @property
    def ndarray(self):
        return self._module.ndarray

    @classmethod
    def is_own_array_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax buffer, otherwise `False`.

        """
        return cls.is_array_type(type_) or cls.is_tracer_type(type_)

    @classmethod
    def is_array_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax buffer, otherwise `False`.

        """
        module, _, suffix = type_.__module__.partition(".")
        return module == "jaxlib"

    @classmethod
    def is_tracer_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax tracer, otherwise `False`.

        """
        module, _, suffix = type_.__module__.partition(".")
        return module == "jax"

    def is_c_contiguous(self, x: ArrayLike) -> bool:
        return True

    def ascontiguousarray(self, x: ArrayLike) -> ArrayLike:
        return x

    def strides(self, x: ArrayLike) -> tuple[int, ...]:
        out: tuple[int, ...] = (x.dtype.itemsize,)
        for item in cast(tuple[int, ...], x.shape[-1:0:-1]):
            out = (item * out[0], *out)
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np  # noqa: F401
import pytest

import awkward as ak

to_list = ak.operations.to_list


def test_broadcast_arrays():
    a = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)
    b = ak.Array([100, 200, 300], check_valid=True)

    out = ak.operations.broadcast_arrays(a, b)
    assert to_list(out[0]) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert to_list(out[1]) == [[100, 100, 100], [], [300, 300]]


numexpr = pytest.importorskip("numexpr")


def test_numexpr():
    # NumExpr's interface pulls variables from the surrounding scope,
    # so these F841 "unused variables" actually are used.

    a = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)
    b = ak.Array([100, 200, 300], check_valid=True)  # noqa: F841
    assert to_list(ak._connect.numexpr.evaluate("a + b")) == [
        [101.1, 102.2, 103.3],
        [],
        [304.4, 305.5],
    ]
    a = [1, 2, 3]  # noqa: F841
    assert to_list(ak._connect.numexpr.re_evaluate()) == [101, 202, 303]



# AUTO GENERATED ON 2023-11-13 AT 18:13:22
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_1():
    index = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops_in = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [2, 0, 2]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [3, 2, 4]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_2():
    index = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops_in = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [8, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_3():
    index = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 4, 5]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_4():
    index = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 7, 6]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 9, 6]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_5():
    index = cupy.array([1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 1, 0, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops_in = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [0, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 1, 1]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_6():
    index = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops_in = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [2, 0, 2]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [3, 2, 4]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_7():
    index = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops_in = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [8, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_8():
    index = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 4, 5]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_9():
    index = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 7, 6]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 9, 6]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_10():
    index = cupy.array([1, 2, 2, 3, 0, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops_in = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [0, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 1, 1]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_11():
    index = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops_in = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [2, 0, 2]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [3, 2, 4]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_12():
    index = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops_in = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [8, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_13():
    index = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 4, 5]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_14():
    index = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 7, 6]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 9, 6]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_15():
    index = cupy.array([1, 3, 0, 3, 5, 2, 0, 2, 1, 1], dtype=cupy.int32)
    starts_in = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops_in = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [0, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 1, 1]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_16():
    index = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    starts_in = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops_in = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [2, 0, 2]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [3, 2, 4]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_17():
    index = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    starts_in = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops_in = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [8, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_18():
    index = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    starts_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 4, 5]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_19():
    index = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    starts_in = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 7, 6]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 9, 6]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_20():
    index = cupy.array([1, 4, 2, 3, 1, 2, 3, 1, 4, 3, 2, 1, 3, 2, 4, 5, 1, 2, 3, 4, 5], dtype=cupy.int32)
    starts_in = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops_in = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [0, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 1, 1]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_21():
    index = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    starts_in = cupy.array([2, 0, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=cupy.int64)
    stops_in = cupy.array([3, 2, 4, 5, 3, 4, 2, 5, 3, 4, 6, 11], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [2, 0, 2]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [3, 2, 4]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_22():
    index = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    starts_in = cupy.array([1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1], dtype=cupy.int64)
    stops_in = cupy.array([8, 4, 5, 6, 5, 5, 7], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [8, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_23():
    index = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    starts_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 4, 5, 6, 5, 5, 7, 1, 2, 1, 3, 1, 5, 3, 2], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 4, 5]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 4, 5]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_24():
    index = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    starts_in = cupy.array([1, 7, 6, 1, 3, 4, 2, 5, 2, 3, 1, 2, 3, 4, 5, 6, 7, 1, 2], dtype=cupy.int64)
    stops_in = cupy.array([1, 9, 6, 2, 4, 5, 3, 6, 3, 4, 2, 4, 5, 5, 7, 8, 2, 3], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [1, 7, 6]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 9, 6]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

def test_cuda_awkward_MaskedArray32_getitem_next_jagged_project_25():
    index = cupy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int32)
    starts_in = cupy.array([0, 0, 0, 0, 0, 0, 0, 0], dtype=cupy.int64)
    stops_in = cupy.array([1, 1, 1, 1, 1, 1, 1, 1], dtype=cupy.int64)
    starts_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    stops_out = cupy.array([123, 123, 123], dtype=cupy.int64)
    length = 3
    funcC = cupy_backend['awkward_MaskedArray_getitem_next_jagged_project', cupy.int32, cupy.int64, cupy.int64, cupy.int64, cupy.int64]
    funcC(index, starts_in, stops_in, starts_out, stops_out, length)

    try:
        ak_cu.synchronize_cuda()
    except:
        pytest.fail("This test case shouldn't have raised an error")
    pytest_starts_out = [0, 0, 0]
    assert cupy.array_equal(starts_out[:len(pytest_starts_out)], cupy.array(pytest_starts_out))
    pytest_stops_out = [1, 1, 1]
    assert cupy.array_equal(stops_out[:len(pytest_stops_out)], cupy.array(pytest_stops_out))

